unit UEngine;

{$mode delphi}

interface

uses
  Classes, SysUtils, Contnrs, Graphics, Controls, ExtCtrls, Math, DateUtils;

type
  TStationShape = (ssCircle, ssSquare, ssTriangle, ssStar, ssPlus, ssPentagon,
    ssDiamond);
  TStationShapeSet = set of TStationShape;
  TEngine = class;
  TMetroPassengers = class;
  TMetroLines = class;
  TMetroLine = class;
  TMetroTrains = class;
  TTrackPoint = class;
  TLineStation = class;

  { TMapStation }

  TMapStation = class
  private
    procedure ShiftTrackPoints;
    procedure SortLines;
  public
    Engine: TEngine;
    Shape: TStationShape;
    Position: TPoint;
    Passengers: TMetroPassengers;
    Lines: TMetroLines;
    ShapeDistance: array[TStationShape] of Integer;
    OverloadDuration: TDateTime;
    function IsBestStationForShape(Shape: TStationShape; Check, Current: TLineStation): Boolean;
    constructor Create;
    destructor Destroy; override;
  end;

  { TMapStations }

  TMapStations = class(TObjectList)
    Engine: TEngine;
    function AddNew: TMapStation;
  end;

  TLineStation = class
    Line: TMetroLine;
    MapStation: TMapStation;
    TrackPoint: TTrackPoint;
  end;

  { TLineStations }

  TLineStations = class(TObjectList)
    Line: TMetroLine;
    function SearchMapStation(Station: TMapStation): TLineStation;
  end;

  { TTrackPoint }

  TTrackPoint = class
    Line: TMetroLine;
    LineStation: TLineStation;
    Point: TPoint;
    PointShift: TPoint;
    DesignedPoint: TPoint;
    Pending: Boolean;
    function GetDown: TTrackPoint;
    function GetUp: TTrackPoint;
  end;

  TTrackPoints = class(TObjectList)
    Line: TMetroLine;
  end;

  TTrack = class
    PointDown: TTrackPoint;
    PointUp: TTrackPoint;
    Line: TMetroLine;
    Shift: TPoint;
  end;

  { TTracks }

  TTracks = class(TObjectList)
    function SearchPointUp(TrackPoint: TTrackPoint; Skip: TTrack): TTrack;
    function SearchPointDown(TrackPoint: TTrackPoint; Skip: TTrack): TTrack;
  end;

  { TTrackPointsAngle }

  TTrackPointsAngle = class
    Angle: Double;
    Tracks: TTracks;
    constructor Create;
    destructor Destroy; override;
  end;

  { TTrackPointsAngleGroup }

  TTrackPointsAngleGroup = class(TObjectList)
    function SearchAngle(Angle: Double): TTrackPointsAngle;
  end;

  { TMetroLine }

  TMetroLine = class
  private
    procedure UpdateEndingLines;
  public
    Index: Integer;
    Engine: TEngine;
    Color: TColor;
    LineStations: TLineStations;
    Trains: TMetroTrains;
    TrackPoints: TTrackPoints;
    procedure ConnectStation(Station: TMapStation; LineStationDown, LineStationUp: TLineStation);
    procedure DisconnectStation(LineStation: TLineStation);
    procedure RouteTrack(TP1, TP2: TTrackPoint);
    procedure RemoteTrackBetween(TP1, TP2: TTrackPoint);
    function GetTrackLength: Integer;
    function GetStationTrackPos(LineStation: TLineStation): Integer;
    constructor Create;
    destructor Destroy; override;
    function IsCircular: Boolean;
  end;

  { TMetroLines }

  TMetroLines = class(TObjectList)
    Engine: TEngine;
    function AddNew: TMetroLine;
    function SearchByColor(Color: TColor): TMetroLine;
  end;

  { TMetroTrain }

  TMetroTrain = class
  private
    LastPosDelta: Integer;
  public
    Passengers: TMetroPassengers;
    Line: TMetroLine;
    TrackPos: Integer;
    //Angle: Double;
    Direction: Integer;
    InStation: Boolean;
    StationStopTime: TDateTime;
    LastTrainMoveTime: TDateTime;
    TargetStation: TLineStation;
    function GetPosition: TPoint;
    function GetAngle: Double;
    constructor Create;
    destructor Destroy; override;
  end;

  { TMetroTrains }

  TMetroTrains = class(TObjectList)
    function GetUnusedTrain: TMetroTrain;
    function GetUnusedCount: Integer;
    function AddNew: TMetroTrain;
  end;

  TMetroPassenger = class
    Engine: TEngine;
    Shape: TStationShape;
    Station: TMapStation;
    Train: TMetroTrain;
  end;

  { TMetroPassengers }

  TMetroPassengers = class(TObjectList)
    Engine: TEngine;
    function AddNew: TMetroPassenger;
  end;

  TMap = class
    Size: TPoint;
  end;

  TView = class
    Size: TPoint;
  end;

  TGameState = (gsNotStarted, gsRunning, gsPaused, gsGameOver);

  { TEngine }

  TEngine = class
  private
    LastMousePos: TPoint;
    LastFocusedStation: TMapStation;
    MouseHold: Boolean;
    LastNewStationTime: TDateTime;
    LastNewPassengerTime: TDateTime;
    LastNewWeekTime: TDateTime;
    LastNewShapeTime: TDateTime;
    LastTickTime: TDateTime;
    ImagePassenger: TImage;
    ImageLocomotive: TImage;
    FTime: TDateTime;
    FLastTime: TDateTime;
    function GetExistStationShapes: TStationShapeSet;
    function GetStationOnPos(Pos: TPoint): TMapStation;
    function GetTrackOnPos(Pos: TPoint): TTrack;
    procedure DrawLine(Canvas: TCanvas; Pos: TPoint);
    procedure DrawShape(Canvas: TCanvas; Position: TPoint; Shape: TStationShape;
      Size: Integer; Angle: Double);
    procedure DrawClock(Canvas: TCanvas);
    procedure DrawTrains(Canvas: TCanvas);
    procedure ComputeShapeDistance;
    procedure ComputeShapeDistanceStation(Station: TMapStation;
      UpdatedShape: TStationShape; Distance: Integer);
    procedure TrainMovement;
    function GetUnusedLine: TMetroLine;
    procedure ShiftTrackPoints;
  public
    Passengers: TMetroPassengers;
    Stations: TMapStations;
    Lines: TMetroLines;
    Trains: TMetroTrains;
    ShapeCount: Integer;
    Map: TMap;
    View: TView;
    SelectedLine: TMetroLine;
    TrackStationDown: TTrackPoint;
    TrackStationUp: TTrackPoint;
    ServedPassengerCount: Integer;
    State: TGameState;
    procedure Tick;
    procedure MouseMove(Position: TPoint);
    procedure MouseUp(Button: TMouseButton; Position: TPoint);
    procedure MouseDown(Button: TMouseButton; Position: TPoint);
    procedure Reset;
    constructor Create;
    destructor Destroy; override;
    procedure Paint(Canvas: TCanvas);
    property Time: TDateTime read FTime;
  end;

const
  clDarkYellow = TColor($00dede);
  clOrange = TColor($0080ff);
  clBrown = TColor($003090);
  LineColors: array[0..7] of TColor = (clBlue, clRed, clDarkYellow, clGreen,
    clPurple, clGray, clOrange, clBrown);
  StationSize = 30;
  StationOverloadSize = 60;
  PassengerSize = 15;
  TrainSize = 40;
  LineColorsDist = 50;
  TrainSpeed = 2000;
  ImagePassengerName = 'Images/Passenger.png';
  ImageLocomotiveName = 'Images/Locomotive.png';
  TrainPassengerCount = 6;
  StationMinDistance = 100;
  StationMaxDistance = 300;
  MaxWaitingPassengers = 10;
  MaxPassengersOveloadTime = 2;
  MetroLineThickness = 13;
  TrackClickDistance = 20;
  EndStationLength = 50;
  ShowDistances = False;
  //TimePerSecond = (60 * OneMinute);
  TimePerSecond = (60 * OneMinute);
  NewStationPeriod = 1;
  NewShapePeriod = 10;
  NewTrainPeriod = 7; // Each week
  NewPassengerPeriod = 0.3 * OneSecond;
  NewPassengerProbability = 0.002;
  VisiblePassengersPerLine = 6;

implementation

uses
  UGeometric;

{ TTracks }

function TTracks.SearchPointUp(TrackPoint: TTrackPoint; Skip: TTrack): TTrack;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TTrack(Items[I]).PointUp <> TrackPoint) and (TTrack(Items[I]) <> Skip) do Inc(I);
  if I < Count then Result := TTrack(Items[I])
    else Result := nil;
end;

function TTracks.SearchPointDown(TrackPoint: TTrackPoint; Skip: TTrack): TTrack;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TTrack(Items[I]).PointDown <> TrackPoint) and (TTrack(Items[I]) <> Skip) do Inc(I);
  if I < Count then Result := TTrack(Items[I])
    else Result := nil;
end;

{ TTrackPointsAngleGroup }

function TTrackPointsAngleGroup.SearchAngle(Angle: Double): TTrackPointsAngle;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TTrackPointsAngle(Items[I]).Angle <> Angle) do Inc(I);
  if I < Count then Result := TTrackPointsAngle(Items[I])
    else Result := nil;
end;

{ TTrackPointsAngle }

constructor TTrackPointsAngle.Create;
begin
  Tracks := TTracks.Create;
  Tracks.OwnsObjects := False;
end;

destructor TTrackPointsAngle.Destroy;
begin
  Tracks.Free;
  inherited Destroy;
end;

{ TTrackPoint }

function TTrackPoint.GetDown: TTrackPoint;
var
  NewIndex: Integer;
begin
  NewIndex := LineStation.Line.LineStations.IndexOf(LineStation) - 1;
  if NewIndex >= 0 then Result := TLineStation(LineStation.Line.LineStations[NewIndex]).TrackPoint
    else Result := nil;
end;

function TTrackPoint.GetUp: TTrackPoint;
var
  NewIndex: Integer;
begin
  NewIndex := LineStation.Line.LineStations.IndexOf(LineStation) + 1;
  if NewIndex < LineStation.Line.LineStations.Count then
    Result := TLineStation(LineStation.Line.LineStations[NewIndex]).TrackPoint
    else Result := nil;
end;

{ TLineStations }

function TLineStations.SearchMapStation(Station: TMapStation): TLineStation;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TLineStation(Items[I]).MapStation <> Station) do Inc(I);
  if I < Count then Result := TLineStation(Items[I])
    else Result := nil;
end;

{ TMetroPassengers }

function TMetroPassengers.AddNew: TMetroPassenger;
begin
  Result := TMetroPassenger.Create;
  Result.Engine := Engine;
  Result.Shape := TStationShape(Random(Integer(Engine.ShapeCount)));
  Add(Result);
end;

{ TMetroTrains }

function TMetroTrains.GetUnusedTrain: TMetroTrain;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (Assigned(TMetroTrain(Items[I]).Line)) do Inc(I);
  if I < Count then Result := TMetroTrain(Items[I])
    else Result := nil;
end;

function TMetroTrains.GetUnusedCount: Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to Count - 1 do
    if not Assigned(TMetroTrain(Items[I]).Line) then Inc(Result);
end;

function TMetroTrains.AddNew: TMetroTrain;
begin
  Result := TMetroTrain.Create;
  Add(Result);
end;

{ TMapStations }

function TMapStations.AddNew: TMapStation;
var
  D: Integer;
  MinD: Integer;
  I: Integer;
  Pass: Integer;
  Angle: Double;
  L: Integer;
const
  Step = 20;
begin
  Result := TMapStation.Create;
  Result.Engine := Engine;
  Angle := Random * 2 * Pi;
  // Ensure minimum distance between stations
  Pass := 0;
  L := Step;
  repeat
    Result.Position := Point(Trunc(Engine.Map.Size.X / 2 + Cos(Angle) * L),
      Trunc(Engine.Map.Size.Y / 2 + Sin(Angle) * L));
    MinD := High(Integer);
    for I := 0 to Engine.Stations.Count - 1 do begin
      D := Distance(TMapStation(Engine.Stations[I]).Position, Result.Position);
      if D < MinD then MinD := D;
    end;
    Inc(Pass);
    L := L + StationMinDistance div 2;
  until (MinD > StationMinDistance) or
    (Pass > 1000) or (Engine.Stations.Count = 0);
  Result.Shape := TStationShape(Random(Integer(Engine.ShapeCount)));
  Add(Result);
  Engine.ComputeShapeDistance;
end;

{ TMetroLines }

function TMetroLines.AddNew: TMetroLine;
begin
  Result := TMetroLine.Create;
  Result.Color := LineColors[Count];
  Result.Engine := Engine;
  Result.Index := Count;
  Add(Result);
end;

function TMetroLines.SearchByColor(Color: TColor): TMetroLine;
var
  I: Integer;
begin
  I := 0;
  while (I < Count) and (TMetroLine(Items[I]).Color <> Color) do Inc(I);
  if I < Count then Result := TMetroLine(Items[I])
    else Result := nil;
end;

{ TMetroLine }

procedure TMetroLine.UpdateEndingLines;
var
  Index: Integer;
  NewTrackPoint: TTrackPoint;
  Angle: Double;
  EndPoint: TPoint;
begin
  if LineStations.Count >= 2 then begin
    Index := TrackPoints.IndexOf(TLineStation(LineStations.First).TrackPoint);
    if Index = 0 then begin
      NewTrackPoint := TTrackPoint.Create;
      NewTrackPoint.Line := Self;
      TrackPoints.Insert(0, NewTrackPoint);
    end;
    Index := TrackPoints.IndexOf(TLineStation(LineStations.Last).TrackPoint);
    if Index = TrackPoints.Count - 1 then begin
      NewTrackPoint := TTrackPoint.Create;
      NewTrackPoint.Line := Self;
      TrackPoints.Insert(TrackPoints.Count, NewTrackPoint);
    end;

    Angle := arctan2((TTrackPoint(TrackPoints[2]).Point.Y - TTrackPoint(TrackPoints[1]).Point.Y),
      (TTrackPoint(TrackPoints[2]).Point.X - TTrackPoint(TrackPoints[1]).Point.X));
    EndPoint := Point(Round(TTrackPoint(TrackPoints[1]).Point.X - EndStationLength * Cos(Angle)),
      Round(TTrackPoint(TrackPoints[1]).Point.Y - EndStationLength * Sin(Angle)));
    TTrackPoint(TrackPoints.First).Point := EndPoint;
    TTrackPoint(TrackPoints.First).DesignedPoint := EndPoint;

    Angle := arctan2((TTrackPoint(TrackPoints[TrackPoints.Count - 2]).Point.Y - TTrackPoint(TrackPoints[TrackPoints.Count - 3]).Point.Y),
      (TTrackPoint(TrackPoints[TrackPoints.Count - 2]).Point.X - TTrackPoint(TrackPoints[TrackPoints.Count - 3]).Point.X));
    EndPoint := Point(Round(TTrackPoint(TrackPoints[TrackPoints.Count - 2]).Point.X + EndStationLength * Cos(Angle)),
      Round(TTrackPoint(TrackPoints[TrackPoints.Count - 2]).Point.Y + EndStationLength * Sin(Angle)));
    TTrackPoint(TrackPoints.Last).Point := EndPoint;
    TTrackPoint(TrackPoints.Last).DesignedPoint := EndPoint;
  end;
end;

procedure TMetroLine.ConnectStation(Station: TMapStation; LineStationDown, LineStationUp: TLineStation);
var
  Train: TMetroTrain;
  NewTrackPoint: TTrackPoint;
  NewLineStation: TLineStation;
  Index: Integer;
begin
  if not Assigned(Station) then
    raise Exception.Create('Station have to be defined');
  if not Assigned(LineStationDown) and not Assigned(LineStationUp) and (LineStations.Count > 0) then
    raise Exception.Create('No old line station to connect new station');
  NewLineStation := TLineStation.Create;
  NewLineStation.Line := Self;
  NewLineStation.MapStation := Station;
  Index := 0;
  if Assigned(LineStationDown) then Index := LineStations.IndexOf(LineStationDown) + 1
    else if Assigned(LineStationDown) then Index := LineStations.IndexOf(LineStationUp);
  LineStations.Insert(Index, NewLineStation);
  Station.Lines.Add(Self);

  NewTrackPoint := TTrackPoint.Create;
  NewTrackPoint.LineStation := NewLineStation;
  NewTrackPoint.Point := Station.Position;
  NewTrackPoint.DesignedPoint := NewTrackPoint.Point;
  NewTrackPoint.Line := TrackPoints.Line;
  Index := 0;
  if Assigned(LineStationDown) then Index := TrackPoints.IndexOf(LineStationDown.TrackPoint) + 1
    else if Assigned(LineStationUp) then Index := TrackPoints.IndexOf(LineStationUp.TrackPoint);
  TrackPoints.Insert(Index, NewTrackPoint);
  NewLineStation.TrackPoint := NewTrackPoint;

  if Assigned(LineStationDown) then
    RouteTrack(NewLineStation.TrackPoint.GetDown, NewLineStation.TrackPoint);
  if Assigned(LineStationUp) then
    RouteTrack(NewLineStation.TrackPoint, NewLineStation.TrackPoint.GetUp);

  // Place one train if at least two stations present
  if (LineStations.Count = 2) then begin
    Train := Engine.Trains.GetUnusedTrain;
    if Assigned(Train) then begin
      Train.Line := Self;
      Train.TargetStation := TLineStation(LineStations[0]);
      Trains.Add(Train);
    end;
  end;
  UpdateEndingLines;
  Engine.ComputeShapeDistance;
  Engine.ShiftTrackPoints;
end;

procedure TMetroLine.DisconnectStation(LineStation: TLineStation);
var
  I: Integer;
  Index: Integer;
begin
  // Delete old trackpoints

  Index := TrackPoints.IndexOf(LineStation.TrackPoint) - 1;
  while (Index >= 0) and (not Assigned(TTrackPoint(TrackPoints[Index]).LineStation)) do begin
    TrackPoints.Delete(Index);
    Dec(Index);
  end;
  Index := Index + 1;
  TrackPoints.Delete(Index);
  while (Index < TrackPoints.Count) and (not Assigned(TTrackPoint(TrackPoints[Index]).LineStation)) do
    TrackPoints.Delete(Index);

  LineStation.MapStation.Lines.Remove(Self);
  Index := LineStations.IndexOf(LineStation);

  // Remove station from trains
  for I := 0 to Trains.Count - 1 do
  with TMetroTrain(Trains[I]) do
  if TargetStation = LineStation then
    TargetStation := TLineStation(LineStations[(Index + 1) mod LineStations.Count]);

  LineStations.Delete(Index);

  // Remove trains if less then two stations
  if LineStations.Count < 2 then
  for I := Trains.Count - 1 downto 0 do begin
    TMetroTrain(Trains[I]).Line := nil;
    Trains.Delete(I);
  end;
  UpdateEndingLines;
  Engine.ComputeShapeDistance;
  Engine.ShiftTrackPoints;
end;

procedure TMetroLine.RouteTrack(TP1, TP2: TTrackPoint);
var
  NewTrackPoint: TTrackPoint;
  Delta: TPoint;
  P1, P2: TPoint;
  Index1, Index2: Integer;
begin
  RemoteTrackBetween(TP1, TP2);
  Index1 := TrackPoints.IndexOf(TP1);
  Index2 := TrackPoints.IndexOf(TP2);
  P1 := TTrackPoint(TrackPoints[Index1]).Point;
  P2 := TTrackPoint(TrackPoints[Index2]).Point;
  NewTrackPoint := TTrackPoint.Create;
  NewTrackPoint.Line := Self;
  Delta := Point(P2.X - P1.X, P2.Y - P1.Y);
  if Abs(Delta.X) > Abs(Delta.Y) then begin
    NewTrackPoint.Point := Point(P2.X - Sign(Delta.X) * Abs(Delta.Y), P1.Y);
    NewTrackPoint.DesignedPoint := NewTrackPoint.Point;
  end else begin
    NewTrackPoint.Point := Point(P1.X, P2.Y - Sign(Delta.Y) * Abs(Delta.X));
    NewTrackPoint.DesignedPoint := NewTrackPoint.Point;
  end;
  TrackPoints.Insert(Index1 + 1, NewTrackPoint);
end;

procedure TMetroLine.RemoteTrackBetween(TP1, TP2: TTrackPoint);
var
  Index1, Index2: Integer;
  Temp: Integer;
  I: Integer;
begin
  Index1 := TrackPoints.IndexOf(TP1);
  Index2 := TrackPoints.IndexOf(TP2);
  if (Index1 = -1) then
    raise Exception.Create('TrackPoint1 not found');
  if (Index2 = -1) then
    raise Exception.Create('TrackPoint2 not found');
  if Index1 > Index2 then begin
    Temp := Index1;
    Index1 := Index2;
    Index2 := Temp;
  end;
  for I := 1 to Index2 - Index1 - 1 do
    TrackPoints.Delete(Index1 + 1);
end;

function TMetroLine.GetTrackLength: Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to TrackPoints.Count - 1 do
  if I > 0 then
    Result := Result + Distance(TTrackPoint(TrackPoints[I]).Point, TTrackPoint(TrackPoints[I - 1]).Point);
end;

function TMetroLine.GetStationTrackPos(LineStation: TLineStation): Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to TrackPoints.Count - 1 do begin
    if I > 0 then
      Result := Result + Distance(TTrackPoint(TrackPoints[I]).Point, TTrackPoint(TrackPoints[I - 1]).Point);
    if TTrackPoint(TrackPoints[I]).LineStation = LineStation then Break;
  end;
end;

constructor TMetroLine.Create;
begin
  LineStations := TLineStations.Create;
  LineStations.OwnsObjects := True;
  Trains := TMetroTrains.Create;
  Trains.OwnsObjects := False;
  TrackPoints := TTrackPoints.Create;
  TrackPoints.Line := Self;
end;

destructor TMetroLine.Destroy;
begin
  TrackPoints.Free;
  Trains.Free;
  LineStations.Free;
  inherited Destroy;
end;

function TMetroLine.IsCircular: Boolean;
begin
  Result := False;
  if LineStations.Count >= 2 then
    Result := (TLineStation(LineStations.Last).MapStation = TLineStation(LineStations.First).MapStation);
end;

{ TMetroTrain }

function TMetroTrain.GetPosition: TPoint;
var
  I: Integer;
  R: TPoint;
  D: Integer;
  Sum: Integer;
begin
  Result := Point(0, 0);
  if Assigned(Line) then
  with Line do begin
    Sum := 0;
    if TrackPoints.Count > 1 then
    for I := 1 to TrackPoints.Count - 1 do begin
      D := Distance(TTrackPoint(TrackPoints[I]).Point, TTrackPoint(TrackPoints[I - 1]).Point);
      if (Sum + D) > Self.TrackPos then begin
        R := Point(TTrackPoint(TrackPoints[I]).Point.X - TTrackPoint(TrackPoints[I - 1]).Point.X,
          TTrackPoint(TrackPoints[I]).Point.Y - TTrackPoint(TrackPoints[I - 1]).Point.Y);
        Result := Point(Trunc(TTrackPoint(TrackPoints[I - 1]).Point.X + R.X * (TrackPos - Sum) / D),
          Trunc(TTrackPoint(TrackPoints[I - 1]).Point.Y + R.Y * (TrackPos - Sum) / D));
        Exit;
      end else Sum := Sum + D;
    end;
    if TrackPoints.Count > 0 then
      Result := TTrackPoint(TrackPoints.Last).Point;
  end;
end;

function TMetroTrain.GetAngle: Double;
var
  I: Integer;
  D: Integer;
  Sum: Integer;
begin
  Result := 0;
  if Assigned(Line) then
  with Line do begin
    Sum := 0;
    if TrackPoints.Count > 1 then
    for I := 1 to TrackPoints.Count - 1 do begin
      D := Distance(TTrackPoint(TrackPoints[I]).Point, TTrackPoint(TrackPoints[I - 1]).Point);
      if (Sum + D) > Self.TrackPos then begin
        Result := ArcTan2(TTrackPoint(TrackPoints[I]).Point.Y - TTrackPoint(TrackPoints[I - 1]).Point.Y,
          TTrackPoint(TrackPoints[I]).Point.X - TTrackPoint(TrackPoints[I - 1]).Point.X);
        Exit;
      end else Sum := Sum + D;
    end;
  end;
end;

constructor TMetroTrain.Create;
begin
  Passengers := TMetroPassengers.Create;
  Passengers.OwnsObjects := False;
  Direction := 1;
  Line := nil;
end;

destructor TMetroTrain.Destroy;
begin
  Passengers.Free;
  inherited Destroy;
end;

{ TMapStation }

procedure TMapStation.ShiftTrackPoints;
var
  Tracks: TTracks;
  I: Integer;
  J: Integer;
  Index: Integer;
  TP: TTrackPoint;
  LS: TLineStation;
  Line: TMetroLine;
  Angle: Double;
  TPAngleGroup: TTrackPointsAngleGroup;
  GroupItem: TTrackPointsAngle;
  NewTrack: TTrack;
  HAngle: Double;
  PairTrack: TTrack;
  NewPoint: TPoint;
begin
  Tracks := TTracks.Create;

  // Collect all near track points as track
  SortLines;
  for I := 0 to Lines.Count - 1 do begin
    Line := TMetroLine(Lines[I]);
    LS := Line.LineStations.SearchMapStation(Self);
    TP := LS.TrackPoint;
    Index := Line.TrackPoints.IndexOf(TP);
    if Index > 0 then begin
      NewTrack := TTrack.Create;
      NewTrack.PointDown := TTrackPoint(Line.TrackPoints[Index - 1]);
      NewTrack.PointUp := TTrackPoint(Line.TrackPoints[Index]);
      Tracks.Add(NewTrack);
    end;
    if Index < (Line.TrackPoints.Count - 1) then begin
      NewTrack := TTrack.Create;
      NewTrack.PointDown := TTrackPoint(Line.TrackPoints[Index + 1]);
      NewTrack.PointUp := TTrackPoint(Line.TrackPoints[Index]);
      Tracks.Add(NewTrack);
    end;
  end;

  // Make groups of tracks with same angle
  TPAngleGroup := TTrackPointsAngleGroup.Create;
  for I := 0 to Tracks.Count - 1 do begin
    Angle := ArcTan2(TTrack(Tracks[I]).PointDown.DesignedPoint.Y - Position.Y,
      TTrack(Tracks[I]).PointDown.DesignedPoint.X - Position.X);
    GroupItem := TPAngleGroup.SearchAngle(Angle);
    if not Assigned(GroupItem) then begin
      GroupItem := TTrackPointsAngle.Create;
      GroupItem.Angle := Angle;
      TPAngleGroup.Add(GroupItem);
    end;
    GroupItem.Tracks.Add(TTrack(Tracks[I]))
  end;

  // Shift tracks according number of lines in group
  for I := 0 to TPAngleGroup.Count - 1 do
  with TTrackPointsAngle(TPAngleGroup[I]) do begin
    for J := 0 to Tracks.Count - 1 do
      with TTrack(Tracks[J]) do begin
        // Get orthogonal angle
        HAngle := Angle + Pi / 2;
        if HAngle > Pi then HAngle := Hangle - Pi;
        Shift.X := Trunc(MetroLineThickness * Cos(HAngle) * (J - (Tracks.Count - 1) / 2));
        Shift.Y := Trunc(MetroLineThickness * Sin(HAngle) * (J - (Tracks.Count - 1) / 2));
        PointDown.PointShift := Shift;
        PointUp.PointShift := Shift;
      end;
  end;

  TPAngleGroup.Free;
  Tracks.Free;
end;

function MapStationCompareLine(Item1, Item2: Pointer): Integer;
begin
  if TMetroLine(Item1).Index > TMetroLine(Item2).Index then Result := 1
  else if TMetroLine(Item1).Index < TMetroLine(Item2).Index then Result := -1
  else Result := 0;
end;

procedure TMapStation.SortLines;
begin
  Lines.Sort(MapStationCompareLine);
end;

function TMapStation.IsBestStationForShape(Shape: TStationShape;
  Check, Current: TLineStation): Boolean;
var
  I: Integer;
  T: Integer;
  Distance: Integer;
  StationIndex: Integer;
  DirectionUp: Boolean;
  DirectionDown: Boolean;
  NextStationUp: TLineStation;
  NextStationDown: TLineStation;
  CurrentLineStation: TLineStation;
begin
  Distance := High(Integer);
  for I := 0 to Lines.Count - 1 do
  with TMetroLine(Lines[I]) do begin
    CurrentLineStation := LineStations.SearchMapStation(Current.MapStation);
    StationIndex := LineStations.IndexOf(CurrentLineStation);
    if IsCircular then begin
      DirectionUp := False;
      DirectionDown := False;
      for T := 0 to Trains.Count - 1 do begin
        if TMetroTrain(Trains[T]).Direction = 1 then DirectionUp := True;
        if TMetroTrain(Trains[T]).Direction = -1 then DirectionDown := True;
      end;
      if StationIndex = 0 then
        NextStationDown := TLineStation(LineStations[LineStations.Count - 2])
      else
      if StationIndex > 0 then
        NextStationDown := TLineStation(LineStations[StationIndex - 1]);

      if (StationIndex >= 0) and (StationIndex = LineStations.Count - 1) then
        NextStationUp := TLineStation(LineStations[1])
      else
      if (StationIndex >= 0) and (StationIndex < LineStations.Count - 1) then
        NextStationUp := TLineStation(LineStations[StationIndex + 1]);
    end else begin
      if StationIndex > 0 then begin
        DirectionDown := True;
        NextStationDown := TLineStation(LineStations[StationIndex - 1])
      end else DirectionDown := False;
      if (StationIndex >= 0) and (StationIndex < LineStations.Count - 1) then begin
        DirectionUp := True;
        NextStationUp := TLineStation(LineStations[StationIndex + 1]);
      end else DirectionUp := False;
    end;
    if DirectionDown and (NextStationDown.MapStation.ShapeDistance[Shape] <> -1) and
      (NextStationDown.MapStation.ShapeDistance[Shape] < Distance) then begin
      Distance := NextStationDown.MapStation.ShapeDistance[Shape];
    end;
    if DirectionUp and (NextStationUp.MapStation.ShapeDistance[Shape] <> -1) and
      (NextStationUp.MapStation.ShapeDistance[Shape] < Distance) then begin
      Distance := NextStationUp.MapStation.ShapeDistance[Shape];
    end;
  end;
  Result := (Check.MapStation.ShapeDistance[Shape] <> -1) and (Check.MapStation.ShapeDistance[Shape] <= Distance);
end;

constructor TMapStation.Create;
begin
  Passengers := TMetroPassengers.Create;
  Passengers.OwnsObjects := False;
  Lines := TMetroLines.Create;
  Lines.OwnsObjects := False;
end;

destructor TMapStation.Destroy;
begin
  Lines.Free;
  Passengers.Free;
  inherited Destroy;
end;

{ TEngine }

function TEngine.GetExistStationShapes: TStationShapeSet;
var
  I: Integer;
begin
  Result := [];
  for I := 0 to Stations.Count - 1 do
  with TMapStation(Stations[I]) do begin
    Result := Result + [Shape];
  end;
end;

function TEngine.GetStationOnPos(Pos: TPoint): TMapStation;
var
  I: Integer;
const
  ClickDistance = 30;
begin
  I := 0;
  while (I < Stations.Count) and (Distance(TMapStation(Stations[I]).Position, Pos) > ClickDistance) do Inc(I);
  if I < Stations.Count then Result := TMapStation(Stations[I])
    else Result := nil;
end;

function TEngine.GetTrackOnPos(Pos: TPoint): TTrack;
var
  I: Integer;
  T: Integer;
  D: Integer;
  MinD: Integer;
begin
  Result := TTrack.Create;
  Result.PointDown := nil;
  Result.PointUp := nil;
  I := 0;
  MinD := High(Integer);
  while (I < Lines.Count) do
  with TMetroLine(Lines[I]) do begin
    for T := 1 to TrackPoints.Count - 1 do begin
      D := PointToLineDistance(Pos, TTrackPoint(TrackPoints[T - 1]).Point, TTrackPoint(TrackPoints[T]).Point);
      if (D < MinD) and (D < TrackClickDistance) then begin
        MinD := D;
        Result.PointDown := TTrackPoint(TrackPoints[T - 1]);
        Result.PointUp := TTrackPoint(TrackPoints[T]);
      end;
    end;
    Inc(I);
  end;
end;

procedure TEngine.DrawLine(Canvas: TCanvas; Pos: TPoint);
var
  Delta: TPoint;
begin
  Delta := Point(Pos.X - Canvas.PenPos.X, Pos.Y - Canvas.PenPos.Y);
  if Abs(Delta.X) > Abs(Delta.Y) then begin
    Canvas.LineTo(Pos.X - Sign(Delta.X) * Abs(Delta.Y), Canvas.PenPos.Y);
  end else begin
    Canvas.LineTo(Canvas.PenPos.X, Pos.Y - Sign(Delta.Y) * Abs(Delta.X));
  end;
  Canvas.LineTo(Pos.X, Pos.Y);
end;

procedure TEngine.DrawShape(Canvas: TCanvas; Position: TPoint; Shape: TStationShape;
  Size: Integer; Angle: Double);
var
  Points: array of TPoint;
  I: Integer;
  Angle2: Double;
begin
  case Shape of
    ssSquare: begin
      SetLength(Points, 4);
      Points[0] := Point(Position.X - Size div 2, Position.Y - Size div 2);
      Points[1] := Point(Position.X + Size div 2, Position.Y - Size div 2);
      Points[2] := Point(Position.X + Size div 2, Position.Y + Size div 2);
      Points[3] := Point(Position.X - Size div 2, Position.Y + Size div 2);
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssCircle: Canvas.Ellipse(
      Position.X - Size div 2, Position.Y - Size div 2,
      Position.X + Size div 2, Position.Y + Size div 2);
    ssTriangle: begin
      SetLength(Points, 3);
      Points[0] := Point(Position.X, Position.Y - Size div 2);
      Points[1] := Point(Position.X + Size div 2, Position.Y + Size div 2);
      Points[2] := Point(Position.X - Size div 2, Position.Y + Size div 2);
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssStar: begin
      SetLength(Points, 10);
      for I := 0 to 9 do begin
        Angle2 := I / 10 * 2 * Pi - Pi / 2;
        if (I mod 2) = 0 then
        Points[I] := Point(Round(Position.X + Cos(Angle2) * Size / 2),
          Round(Position.Y + Sin(Angle2) * Size / 2))
        else
        Points[I] := Point(Round(Position.X + Cos(Angle2) * Size / 5),
          Round(Position.Y + Sin(Angle2) * Size / 5));
      end;
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssPlus: begin
      SetLength(Points, 12);
      Points[0] := Point(Position.X + Size div 6, Position.Y - Size div 6);
      Points[1] := Point(Position.X + Size div 2, Position.Y - Size div 6);
      Points[2] := Point(Position.X + Size div 2, Position.Y + Size div 6);
      Points[3] := Point(Position.X + Size div 6, Position.Y + Size div 6);
      Points[4] := Point(Position.X + Size div 6, Position.Y + Size div 2);
      Points[5] := Point(Position.X - Size div 6, Position.Y + Size div 2);
      Points[6] := Point(Position.X - Size div 6, Position.Y + Size div 6);
      Points[7] := Point(Position.X - Size div 2, Position.Y + Size div 6);
      Points[8] := Point(Position.X - Size div 2, Position.Y - Size div 6);
      Points[9] := Point(Position.X - Size div 6, Position.Y - Size div 6);
      Points[10] := Point(Position.X - Size div 6, Position.Y - Size div 2);
      Points[11] := Point(Position.X + Size div 6, Position.Y - Size div 2);
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssPentagon: begin
      SetLength(Points, 5);
      for I := 0 to 4 do begin
        Angle2 := I / 5 * 2 * Pi - Pi / 2;
        Points[I] := Point(Round(Position.X + Cos(Angle2) * Size / 2),
          Round(Position.Y + Sin(Angle2) * Size / 2));
      end;
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
    ssDiamond: begin
      SetLength(Points, 4);
      Points[0] := Point(Position.X, Position.Y - Size div 2);
      Points[1] := Point(Position.X + Size div 2, Position.Y);
      Points[2] := Point(Position.X, Position.Y + Size div 2);
      Points[3] := Point(Position.X - Size div 2, Position.Y);
      Points := RotatePoints(Position, Points, Angle);
      Canvas.Polygon(Points);
    end;
  end;
end;

procedure TEngine.ComputeShapeDistance;
var
  I: Integer;
  L: Integer;
  S: TStationShape;
begin
  // Reset all distances
  for I := 0 to Stations.Count - 1 do
  with TMapStation(Stations[I]) do begin
    for S := Low(ShapeDistance) to High(ShapeDistance) do
      ShapeDistance[S] := -1;
  end;

  // Propagate shape distance for all stations
  // Distace 0 means that station is final target
  for I := 0 to Stations.Count - 1 do
  with TMapStation(Stations[I]) do begin
    ComputeShapeDistanceStation(TMapStation(Stations[I]), Shape, 0);
  end;
end;

procedure TEngine.ComputeShapeDistanceStation(Station: TMapStation;
  UpdatedShape: TStationShape; Distance: Integer);
var
  I: Integer;
  T: Integer;
  StationIndex: Integer;
  DirectionDown: Boolean;
  DirectionUp: Boolean;
begin
  with Station do begin
    if (Distance < ShapeDistance[UpdatedShape]) or (ShapeDistance[UpdatedShape] = -1) then begin
      ShapeDistance[UpdatedShape] := Distance;
      // Do for all lines connected to station
      for I := 0 to Lines.Count - 1 do
      with TMetroLine(Lines[I]) do
        for StationIndex := 0 to LineStations.Count - 1 do
        if TLineStation(LineStations[StationIndex]).MapStation = Station then begin
        if not IsCircular then begin
          // Update for all adjecent stations
          if StationIndex > 0 then
            ComputeShapeDistanceStation(TLineStation(LineStations[StationIndex - 1]).MapStation,
            UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
          if (StationIndex >= 0) and (StationIndex < LineStations.Count - 1) then
            ComputeShapeDistanceStation(TLineStation(LineStations[StationIndex + 1]).MapStation,
            UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
        end else begin
          // If circular then trains might go in single direction so passengers
          // waiting for oposite directions are wrong
          DirectionUp := False;
          DirectionDown := False;
          for T := 0 to Trains.Count - 1 do begin
            if TMetroTrain(Trains[T]).Direction = 1 then DirectionUp := True;
            if TMetroTrain(Trains[T]).Direction = -1 then DirectionDown := True;
          end;
          // Update for all adjecent stations
          if DirectionUp then begin
            if StationIndex = 0 then
              ComputeShapeDistanceStation(TLineStation(LineStations[LineStations.Count - 2]).MapStation,
              UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
            if StationIndex > 0 then
              ComputeShapeDistanceStation(TLineStation(LineStations[StationIndex - 1]).MapStation,
              UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
          end;
          if DirectionDown then begin
            if (StationIndex >= 0) and (StationIndex = LineStations.Count - 1) then
              ComputeShapeDistanceStation(TLineStation(LineStations[1]).MapStation,
              UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
            if (StationIndex >= 0) and (StationIndex < LineStations.Count - 1) then
              ComputeShapeDistanceStation(TLineStation(LineStations[StationIndex + 1]).MapStation,
              UpdatedShape, Station.ShapeDistance[UpdatedShape] + 1);
          end;
        end;
      end;
    end;
  end;
end;

procedure TEngine.TrainMovement;
var
  I: Integer;
  CurrentStation: TLineStation;
  P: Integer;
  Passenger: TMetroPassenger;
  PosDelta: Integer;
  TargetStationIndex: Integer;
  PosChange: Integer;
begin
  // Move trains
  for I := 0 to Trains.Count - 1 do
  with TMetroTrain(Trains[I]) do begin
    if Assigned(Line) then begin
      if InStation then begin
        if (Time - StationStopTime) > OneHour then begin
          CurrentStation := TargetStation;

          // Choose next target station
          TargetStationIndex := Line.LineStations.IndexOf(TargetStation) + Direction;
          if TargetStationIndex < 0 then begin
            if Line.IsCircular then begin
              TargetStationIndex := Line.LineStations.Count - 2;
              TrackPos := Line.GetStationTrackPos(TLineStation(Line.LineStations.Last));
            end else begin
              TargetStationIndex := 1;
              Direction := -Direction;
            end;
          end else if TargetStationIndex >= Line.LineStations.Count then begin
            if Line.IsCircular then begin
              TargetStationIndex := 1;
              TrackPos := 0;
            end else begin
              TargetStationIndex := Line.LineStations.Count - 2;
              Direction := -Direction;
            end;
          end;
          TargetStation := TLineStation(Line.LineStations[TargetStationIndex]);

          // Unload passengers in target station
          if Assigned(CurrentStation) then
          for P := Passengers.Count - 1 downto 0 do begin
            if TMetroPassenger(Passengers[P]).Shape = CurrentStation.MapStation.Shape then begin
              Passenger := TMetroPassenger(Passengers[P]);
              Passengers.Delete(P);
              Self.Passengers.Remove(Passenger);
              Inc(ServedPassengerCount);
            end;
          end;
          // Unload passengers to change line
          if Assigned(CurrentStation)  then
          for P := Passengers.Count - 1 downto 0 do begin
            if not CurrentStation.MapStation.IsBestStationForShape(TMetroPassenger(Passengers[P]).Shape,
            TargetStation, CurrentStation) then begin
              Passenger := TMetroPassenger(Passengers[P]);
              Passengers.Delete(P);
              CurrentStation.MapStation.Passengers.Add(Passenger);
              Passenger.Station := CurrentStation.MapStation;
            end;
          end;

          // Load new passengers
          if Assigned(CurrentStation) and not Assigned(CurrentStation.MapStation) then
            raise Exception.Create('Station have to have MapStation');
          if Assigned(CurrentStation)  then
          for P := CurrentStation.MapStation.Passengers.Count - 1 downto 0 do begin
            if (Passengers.Count < TrainPassengerCount) then begin
              Passenger := TMetroPassenger(CurrentStation.MapStation.Passengers[P]);
              if CurrentStation.MapStation.IsBestStationForShape(Passenger.Shape,
            TargetStation, CurrentStation) then begin
                Passenger.Station := nil;
                CurrentStation.MapStation.Passengers.Delete(P);
                Passengers.Add(Passenger);
                Passenger.Train := TMetroTrain(Trains[I]);
              end;
            end else Break; // No more space
          end;

          LastPosDelta := Abs(TrackPos - Line.GetStationTrackPos(TargetStation));
          InStation := False;
          LastTrainMoveTime := Time;
        end;
      end else begin
        PosChange := Trunc(Direction * TrainSpeed * (Time - LastTrainMoveTime));
        TrackPos := TrackPos + PosChange;
        LastTrainMoveTime := Time;
        if TrackPos < 0 then begin
          if Line.IsCircular then TrackPos := Line.GetTrackLength
            else TrackPos := 0;
        end else
        if TrackPos > Line.GetTrackLength then begin
          if Line.IsCircular then TrackPos := 0
            else TrackPos := Line.GetTrackLength;
        end;
        PosDelta := Abs(TrackPos - Line.GetStationTrackPos(TargetStation));
        if PosDelta >= LastPosDelta then begin
          // We are getting far from station, stop at station
          TrackPos := Line.GetStationTrackPos(TargetStation);
          InStation := True;
          StationStopTime := Time;
        end;
        LastPosDelta := PosDelta;
      end;
    end;
  end;
end;

function TEngine.GetUnusedLine: TMetroLine;
var
  I: Integer;
begin
  I := 0;
  while (I < Lines.Count) and (TMetroLine(Lines[I]).TrackPoints.Count > 0) do Inc(I);
  if I < Lines.Count then Result := TMetroLine(Lines[I])
    else Result := nil;
end;

procedure TEngine.ShiftTrackPoints;
var
  I: Integer;
  J: Integer;
  L: Integer;
  Link1, Link2: TPoint;
  NewPoint: TPoint;
begin
  // Reset all trackpoints position shift
  for I := 0 to Lines.Count - 1 do
  with TMetroLine(Lines[I]) do
    for J := 0 to TrackPoints.Count - 1 do
    TTrackPoint(TrackPoints[J]).Point := TTrackPoint(TrackPoints[J]).DesignedPoint;

  // Calculate new position shifts
  for I := 0 to Stations.Count - 1 do
    TMapStation(Stations[I]).ShiftTrackPoints;

  // Compute track points from track shift
  for L := 0 to Lines.Count - 1 do
  with TMetroLine(Lines[L]) do begin
    if TrackPoints.Count > 0 then
      TTrackPoint(TrackPoints[0]).Point := AddPoint(TTrackPoint(TrackPoints[0]).DesignedPoint,
        TTrackPoint(TrackPoints[0]).PointShift);
    for I := 1 to TrackPoints.Count - 1 do
    with TTrackPoint(TrackPoints[I]) do begin
      Link1 := SubPoint(AddPoint(TTrackPoint(TrackPoints[I]).DesignedPoint, TTrackPoint(TrackPoints[I]).PointShift),
        AddPoint(TTrackPoint(TrackPoints[I - 1]).DesignedPoint, TTrackPoint(TrackPoints[I - 1]).PointShift));
      if (I + 1) < TrackPoints.Count then
        Link2 := SubPoint(AddPoint(TTrackPoint(TrackPoints[I + 1]).DesignedPoint, TTrackPoint(TrackPoints[I + 1]).PointShift),
          AddPoint(TTrackPoint(TrackPoints[I]).DesignedPoint, TTrackPoint(TrackPoints[I]).PointShift))
        else Link2 := Link1;

      if ArcTanPoint(Link1) = ArcTanPoint(Link2) then begin
        // Parallel lines
        TTrackPoint(TrackPoints[I]).Point := AddPoint(TTrackPoint(TrackPoints[I]).DesignedPoint,
          TTrackPoint(TrackPoints[I]).PointShift);
      end else begin
        // Intersected lines
        NewPoint := LineIntersect(AddPoint(TTrackPoint(TrackPoints[I - 1]).DesignedPoint, TTrackPoint(TrackPoints[I - 1]).PointShift),
          AddPoint(TTrackPoint(TrackPoints[I]).DesignedPoint, TTrackPoint(TrackPoints[I]).PointShift),
          AddPoint(TTrackPoint(TrackPoints[I]).DesignedPoint, TTrackPoint(TrackPoints[I]).PointShift),
          AddPoint(TTrackPoint(TrackPoints[I + 1]).DesignedPoint, TTrackPoint(TrackPoints[I + 1]).PointShift));
        TTrackPoint(TrackPoints[I]).Point := NewPoint;
      end;
    end;
  end;

end;

procedure TEngine.DrawClock(Canvas: TCanvas);
var
  ClockCenter: TPoint;
  Angle: Double;
  Text: string;
  I: Integer;
const
  ClockSize = 20;
begin
  Canvas.Pen.Style := psSolid;
  Canvas.Pen.Color := clBlack;
  Canvas.Pen.Width := 2;
  ClockCenter := Point(Canvas.Width - 30, 40);
  Angle := Time / (12 * OneHour) * 2 * Pi - Pi / 2;
  Canvas.EllipseC(ClockCenter.X, ClockCenter.Y, ClockSize, ClockSize);
  Canvas.Line(ClockCenter, Point(ClockCenter.X + Round(Cos(Angle) * ClockSize * 0.8),
    ClockCenter.Y + Round(Sin(Angle) * ClockSize * 0.8)));
  Text := FormatDateTime('ddd', Time);
  Canvas.TextOut(ClockCenter.X - ClockSize - Canvas.TextWidth(Text) - 5, ClockCenter.Y -
    Canvas.TextWidth(Text) div 2, Text);
  for I := 0 to 12 do begin
    Angle := I / 12 * 2 * Pi;
    Canvas.Line(ClockCenter.X + Round(Cos(Angle) * ClockSize * 0.8),
      ClockCenter.Y + Round(Sin(Angle) * ClockSize * 0.8),
      ClockCenter.X + Round(Cos(Angle) * ClockSize * 0.9),
      ClockCenter.Y + Round(Sin(Angle) * ClockSize * 0.9));
  end;
end;

procedure TEngine.DrawTrains(Canvas: TCanvas);
var
  I: Integer;
  P: Integer;
  Pos: TPoint;
  Points: array of TPoint;
  Angle: Double;
  ShapePos: TPoint;
begin
    // Draw trains
  for I := 0 to Trains.Count - 1 do
  with TMetroTrain(Trains[I]) do begin
    if Assigned(Line) then begin
      Canvas.Brush.Color := Line.Color;
      Canvas.Brush.Style := bsSolid;
      Canvas.Pen.Style := psClear;
      Pos := GetPosition;
      Angle := GetAngle;

      SetLength(Points, 4);
      Points[0] := RotatePoint(Pos, Point(Pos.X - TrainSize div 2, Pos.Y - TrainSize div 3), Angle);
      Points[1] := RotatePoint(Pos, Point(Pos.X + TrainSize div 2, Pos.Y - TrainSize div 3), Angle);
      Points[2] := RotatePoint(Pos, Point(Pos.X + TrainSize div 2, Pos.Y + TrainSize div 3), Angle);
      Points[3] := RotatePoint(Pos, Point(Pos.X - TrainSize div 2, Pos.Y + TrainSize div 3), Angle);
      Canvas.Polygon(Points);
      Canvas.Brush.Color := clWhite;
      for P := 0 to Passengers.Count - 1 do
      with TMetroPassenger(Passengers[P]) do begin
        ShapePos := Point(Pos.X - Trunc(TrainSize div 3 * 1) + (P mod 3) * TrainSize div 3,
          Pos.Y - Trunc(TrainSize div 6 * 1) + (P div 3) * TrainSize div 3);
        ShapePos := RotatePoint(Pos, ShapePos, Angle);
        DrawShape(Canvas, ShapePos, Shape, TrainSize div 3, Angle + Pi / 2);
      end;
    end;
  end;
end;

procedure TEngine.Tick;
var
  Passenger: TMetroPassenger;
  I: Integer;
begin
  if State = gsRunning then begin
    FTime := FTime + (Now - LastTickTime) / OneSecond * TimePerSecond;

  // Add new trains
  if (Time - LastNewWeekTime) > NewTrainPeriod then begin
    LastNewWeekTime := Time;
    Trains.AddNew;
    // TODO: Show notification screen with confirmation
  end;

  // Add new shape
  if (Time - LastNewShapeTime) > NewShapePeriod then begin
    LastNewShapeTime := Time;
    if ShapeCount <= Integer(High(TStationShape)) then Inc(ShapeCount);
  end;

  // Add new stations
  if (Time - LastNewStationTime) > NewStationPeriod then begin
    LastNewStationTime := Time;
    Stations.AddNew;
  end;

  // Add new passengers
  if (Time - LastNewPassengerTime) > NewPassengerPeriod then begin
    LastNewPassengerTime := Time;
    for I := 0 to Stations.Count - 1 do
    with TMapStation(Stations[I]) do
    if Random < NewPassengerProbability then begin
      Passenger := Self.Passengers.AddNew;
      Passenger.Station := TMapStation(Stations[I]);
      Passengers.Add(Passenger);

      // Passenger is not allowed to have same shape
      while (Passenger.Shape = Passenger.Station.Shape) or not (Passenger.Shape in GetExistStationShapes) do
        Passenger.Shape := TStationShape((Integer(Passenger.Shape) + 1) mod Integer(ShapeCount));

    end;
  end;

  for I := 0 to Stations.Count - 1 do
  with TMapStation(Stations[I]) do begin
    if Passengers.Count > MaxWaitingPassengers then begin
      OverloadDuration := OverloadDuration + (FTime - FLastTime);
      if OverloadDuration > MaxPassengersOveloadTime then
        OverloadDuration := MaxPassengersOveloadTime;
    end;
    if Passengers.Count <= MaxWaitingPassengers then begin
      OverloadDuration := OverloadDuration - (FTime - FLastTime);
      if OverloadDuration < 0 then begin
        OverloadDuration := 0;
      end;
    end;
  end;

  TrainMovement;

  // Game over
  for I := 0 to Stations.Count - 1 do
  with TMapStation(Stations[I]) do begin
    if OverloadDuration >= MaxPassengersOveloadTime then State := gsGameOver;
  end;

  end;
  LastTickTime := Now;
  FLastTime := FTime;
end;

procedure TEngine.MouseMove(Position: TPoint);
var
  FocusedStation: TMapStation;
  Line: TMetroLine;
  LineStationDown: TLineStation;
  LineStationUp: TLineStation;
  CurrentTrackPoint: TTrackPoint;
begin
  LastMousePos := Position;
  if MouseHold then begin
      FocusedStation := GetStationOnPos(Position);
      Line := nil;
      if Assigned(TrackStationDown) then Line := TrackStationDown.Line;
      if Assigned(TrackStationUp) then Line := TrackStationUp.Line;
      if Assigned(Line) and not Assigned(LastFocusedStation) and Assigned(FocusedStation) then begin
        if Assigned(TrackStationDown) and (TrackStationDown.LineStation.MapStation = FocusedStation) then begin
          // Disconnect down
          CurrentTrackPoint := TrackStationDown;
          TrackStationDown := TrackStationDown.GetDown;
          Line.DisconnectStation(CurrentTrackPoint.LineStation);
        end else
        if Assigned(TrackStationUp) and (TrackStationUp.LineStation.MapStation = FocusedStation) then begin
          // Disconnect up
          CurrentTrackPoint := TrackStationUp;
          if Assigned(TrackStationUp) then
            TrackStationUp := TrackStationUp.GetUp;
          Line.DisconnectStation(CurrentTrackPoint.LineStation);
        end else
        if Assigned(Line) and ((not Line.IsCircular) or ((TrackStationDown <> nil) and (TrackStationUp <> nil))) and
        ((Line.LineStations.SearchMapStation(FocusedStation) = nil) or
        ((Line.LineStations.Count > 0) and
        ((TLineStation(Line.LineStations.First).MapStation = FocusedStation) or
        (TLineStation(Line.LineStations.Last).MapStation = FocusedStation)) and
        ((TrackStationDown = nil) or (TrackStationUp = nil)) and
        (not Line.IsCircular))) then begin
          if Assigned(TrackStationDown) then LineStationDown := TrackStationDown.LineStation
            else LineStationDown := nil;
          if Assigned(TrackStationUp) then LineStationUp := TrackStationUp.LineStation
            else LineStationUp := nil;
          Line.ConnectStation(FocusedStation, LineStationDown, LineStationUp);
          if Assigned(TrackStationDown) then TrackStationDown := TrackStationDown.GetUp
            else if Assigned(TrackStationUp) then TrackStationUp := TrackStationUp.GetDown;
        end;
      end;
      LastFocusedStation := FocusedStation;
  end;
end;

procedure TEngine.MouseUp(Button: TMouseButton; Position: TPoint);
var
  Station: TMapStation;
  Line: TMetroLine;
  I: Integer;
begin
  if Button = mbLeft then begin

    // Line color selection
    for I := 0 to Lines.Count - 1 do
      if Distance(Point(View.Size.X div 2 - Length(LineColors) div 2 * LineColorsDist + I * LineColorsDist,
        View.Size.Y - LineColorsDist), Position) < 20 then begin
          SelectedLine := TMetroLine(Lines[I]);
          Exit;
        end;

    // Remove single line station on line
    if Assigned(TrackStationDown) and (TrackStationDown.Line.LineStations.Count = 1) then begin
      TrackStationDown.Line.DisconnectStation(TLineStation(TrackStationDown.Line.LineStations.First));
    end;
    if Assigned(TrackStationUp) and (TrackStationUp.Line.LineStations.Count = 1) then begin
      TrackStationUp.Line.DisconnectStation(TLineStation(TrackStationUp.Line.LineStations.First));
    end;
  end else
  if Button = mbRight then begin
    SelectedLine := nil;
  end;
  MouseHold := False;
  TrackStationDown := nil;
  TrackStationUp := nil;
end;

procedure TEngine.MouseDown(Button: TMouseButton; Position: TPoint);
var
  Station: TMapStation;
  Line: TMetroLine;
  I: Integer;
  NewLine: TMetroLine;
  Track: TTrack;
  NewIndex: Integer;
begin
  if Button = mbLeft then begin
    MouseHold := True;
    LastFocusedStation := nil;

    // Line selection
    Track := GetTrackOnPos(Position);
    if Assigned(Track) and Assigned(Track.PointDown) and Assigned(Track.PointUp) then begin
      SelectedLine := Track.PointDown.Line;

      TrackStationDown := Track.PointDown;
      NewIndex := TrackStationDown.Line.TrackPoints.IndexOf(TrackStationDown);
      while Assigned(TrackStationDown) and (not Assigned(TrackStationDown.LineStation)) do begin
        NewIndex := NewIndex - 1;
        if NewIndex >= 0 then TrackStationDown := TTrackPoint(TrackStationDown.Line.TrackPoints[NewIndex])
          else TrackStationDown := nil;
      end;
      TrackStationUp := Track.PointUp;
      NewIndex := TrackStationUp.Line.TrackPoints.IndexOf(TrackStationDown);
      while Assigned(TrackStationUp) and (not Assigned(TrackStationUp.LineStation)) do begin
        NewIndex := NewIndex + 1;
        if NewIndex < TrackStationUp.Line.TrackPoints.Count then
          TrackStationUp := TTrackPoint(TrackStationUp.Line.TrackPoints[NewIndex])
          else TrackStationUp := nil;
      end;
      Track.Free;
      Exit;
    end;

    // New track creation from selected station as start
    Station := GetStationOnPos(Position);
    if Assigned(Station) then begin
      if Assigned(SelectedLine) and (SelectedLine.LineStations.Count = 0) then NewLine := SelectedLine
        else NewLine := GetUnusedLine;
      if Assigned(NewLine) then begin
        NewLine.ConnectStation(Station, nil, nil);
        TrackStationDown := TTrackPoint(NewLine.TrackPoints.Last);
        TrackStationUp := nil;
        LastFocusedStation := Station;
      end;
    end;
  end;
end;

procedure TEngine.Reset;
var
  NewTrain: TMetroTrain;
  I: Integer;
  NewStation: TMapStation;
  InitialStationCount: Integer;
begin
  Passengers.Clear;
  Lines.Clear;
  Stations.Clear;

  ShapeCount := 3;
  ServedPassengerCount := 0;

  // Start with 3 stations with each different shape
  InitialStationCount := 3;
  for I := 0 to InitialStationCount - 1 do begin
    NewStation := Stations.AddNew;
    if I = 0 then NewStation.Shape := ssSquare
    else if I = 1 then NewStation.Shape := ssCircle
    else if I = 2 then NewStation.Shape := ssTriangle;
  end;

  for I := 0 to 7 do begin
    Lines.AddNew;
    NewTrain := TMetroTrain.Create;
    Trains.Add(NewTrain);
  end;

  SelectedLine := nil;
  FTime := 0;
  FLastTime := 0;
  LastNewStationTime := Time;
  LastNewPassengerTime := Time;
  LastNewWeekTime := Time;
  LastNewShapeTime := Time;
  LastTickTime := Now;
  State := gsRunning;
end;

constructor TEngine.Create;
begin
  Stations := TMapStations.Create;
  Stations.Engine := Self;
  Lines := TMetroLines.Create;
  Lines.Engine := Self;
  Passengers := TMetroPassengers.Create;
  Passengers.Engine := Self;
  Map := TMap.Create;
  View := TView.Create;
  Trains := TMetroTrains.Create;
  ImagePassenger := TImage.Create(nil);
  ImageLocomotive := TImage.Create(nil);
  if FileExists(ImagePassengerName) then
    ImagePassenger.Picture.LoadFromFile(ImagePassengerName);
  if FileExists(ImageLocomotiveName) then
    ImageLocomotive.Picture.LoadFromFile(ImageLocomotiveName);
end;

destructor TEngine.Destroy;
begin
  Trains.Free;
  ImageLocomotive.Free;
  ImagePassenger.Free;
  View.Free;
  Map.Free;
  Passengers.Free;
  Stations.Free;
  Lines.Free;
  inherited Destroy;
end;

procedure TEngine.Paint(Canvas: TCanvas);
var
  I: Integer;
  S: Integer;
  Size: Integer;
  P: Integer;
  Pos: TPoint;
  Text: string;
  Angle: Double;
  EndPoint: TPoint;
  PassengerPos: TPoint;
  Direction: Integer;
const
  GameOverText = 'Game Over';
  GameOverReason = 'Overcrowding at this station has forced you to resign as metro manager.';
  GameOverStatistic = '%d passengers travelled on your metro over %d days.';
begin
  Canvas.Brush.Color := $eff0e0;
  Canvas.Brush.Style := bsSolid;
  Canvas.Clear;

  // Draw station passenger overload
  for I := 0 to Stations.Count - 1 do
  with TMapStation(Stations[I]) do begin
    if OverloadDuration > 0 then begin
      Canvas.Brush.Color := clSilver;
      Canvas.Brush.Style := bsSolid;
      Canvas.Pen.Color := clSilver;
      Canvas.Pen.Style := psSolid;
      Angle := OverloadDuration / MaxPassengersOveloadTime * 2 * Pi;
      Canvas.Pie(Position.X - StationOverloadSize, Position.Y - StationOverloadSize,
      Position.X + StationOverloadSize, Position.Y + StationOverloadSize,
        Trunc(Position.X + StationOverloadSize * Cos(Angle)),
        Trunc(Position.Y + StationOverloadSize * Sin(Angle)), Position.X + StationOverloadSize, Position.Y);
    end;
  end;

  // Draw lines
  for I := 0 to Lines.Count - 1 do
  with TMetroLine(Lines[I]) do begin
    Canvas.Pen.Color := Color;
    Canvas.Pen.Style := psSolid;
    Canvas.Pen.Width := MetroLineThickness;
    if TrackPoints.Count > 0 then Canvas.MoveTo(TTrackPoint(TrackPoints[0]).Point);
    for S := 1 to TrackPoints.Count - 1 do begin
      Canvas.LineTo(TTrackPoint(TrackPoints[S]).Point);
{      if (S = TrackPoints.Count - 1) then begin
        Canvas.Pen.EndCap := pecSquare;
        Angle := arctan2((TTrackPoint(TrackPoints[S]).Point.Y - TTrackPoint(TrackPoints[S - 1]).Point.Y),
          (TTrackPoint(TrackPoints[S]).Point.X - TTrackPoint(TrackPoints[S - 1]).Point.X));
        EndPoint := Point(Round(TTrackPoint(TrackPoints[S]).Point.X + EndStationLength * Cos(Angle)),
          Round(TTrackPoint(TrackPoints[S]).Point.Y + EndStationLength * Sin(Angle)));
        Canvas.LineTo(EndPoint);
        Canvas.MoveTo(Point(Round(EndPoint.X + Cos(Angle + Pi / 2) * EndStationLength / 3),
          Round(EndPoint.Y + Sin(Angle + Pi / 2) * EndStationLength / 3)));
        Canvas.LineTo(Point(Round(EndPoint.X + Cos(Angle - Pi / 2) * EndStationLength / 3),
          Round(EndPoint.Y + Sin(Angle - Pi / 2) * EndStationLength / 3)));
        Canvas.Pen.EndCap := pecRound;
      end;}
    end;
    {
    if (TrackPoints.Count > 1) then begin
      Canvas.Pen.EndCap := pecSquare;
      Angle := arctan2((TTrackPoint(TrackPoints[1]).Point.Y - TTrackPoint(TrackPoints[0]).Point.Y),
        (TTrackPoint(TrackPoints[1]).Point.X - TTrackPoint(TrackPoints[0]).Point.X));
      Canvas.MoveTo(TTrackPoint(TrackPoints[0]).Point);
      EndPoint := Point(Round(TTrackPoint(TrackPoints[0]).Point.X - EndStationLength * Cos(Angle)),
        Round(TTrackPoint(TrackPoints[0]).Point.Y - EndStationLength * Sin(Angle)));
      Canvas.LineTo(EndPoint);
      Canvas.MoveTo(Point(Round(EndPoint.X - Cos(Angle + Pi / 2) * EndStationLength / 3),
        Round(EndPoint.Y - Sin(Angle + Pi / 2) * EndStationLength / 3)));
      Canvas.LineTo(Point(Round(EndPoint.X - Cos(Angle - Pi / 2) * EndStationLength / 3),
        Round(EndPoint.Y - Sin(Angle - Pi / 2) * EndStationLength / 3)));
      Canvas.Pen.EndCap := pecRound;
    end;  }
  end;

  // Draw design time lines
  if Assigned(TrackStationDown) and Assigned(TrackStationDown.LineStation) then begin
    Canvas.Pen.Color := TrackStationDown.Line.Color;
    Canvas.MoveTo(TrackStationDown.LineStation.TrackPoint.Point);
    DrawLine(Canvas, LastMousePos);
  end;
  if Assigned(TrackStationUp) and Assigned(TrackStationUp.LineStation) then begin
    Canvas.Pen.Color := TrackStationUp.Line.Color;
    Canvas.MoveTo(TrackStationUp.LineStation.TrackPoint.Point);
    DrawLine(Canvas, LastMousePos);
  end;

  DrawTrains(Canvas);

  // Draw stations
  Canvas.Pen.Width := 5;
  for I := 0 to Stations.Count - 1 do
  with TMapStation(Stations[I]) do begin
    Canvas.Pen.Style := psSolid;
    if Assigned(SelectedLine) and (Lines.IndexOf(SelectedLine) <> -1) then begin
      Canvas.Brush.Style := bsClear;
      Canvas.Pen.Color := SelectedLine.Color;
      DrawShape(Canvas, Position, Shape, StationSize + Canvas.Pen.Width + 6, 0);
    end;

    Canvas.Brush.Color := clWhite;
    Canvas.Brush.Style := bsSolid;
    Canvas.Pen.Color := clBlack;
    DrawShape(Canvas, Position, Shape, StationSize, 0);

    // Draw passengers
    Canvas.Pen.Style := psClear;
    Canvas.Brush.Color := clBlack;
    PassengerPos := Point(0, 0);
    Direction := 1;
    for P := 0 to Passengers.Count - 1 do
    with TMetroPassenger(Passengers[P]) do begin
      DrawShape(Canvas, Point(Position.X + StationSize + PassengerPos.X,
        Position.Y - StationSize div 2 + PassengerPos.Y),
        Shape, PassengerSize, 0);
      PassengerPos := Point(PassengerPos.X + Direction * (PassengerSize + 2), PassengerPos.Y);
      if PassengerPos.X >= (PassengerSize + 2) * VisiblePassengersPerLine then begin
        Direction := -Direction;
        PassengerPos.X := PassengerPos.X - (PassengerSize + 2);
        PassengerPos.Y := PassengerPos.Y + (PassengerSize + 2);
      end;
      if PassengerPos.X < 0 then begin
        Direction := -Direction;
        PassengerPos.X := 0;
        PassengerPos.Y := PassengerPos.Y + (PassengerSize + 2);
      end;
    end;

    if ShowDistances then begin
      Canvas.Brush.Style := bsClear;
      Text := '';
      for P := 0 to 5 do
        Text := Text + IntToStr(ShapeDistance[TStationShape(P)]) + ',';
      Canvas.TextOut(Position.X + StationSize div 2, Position.Y + StationSize div 2, Text);
    end;
  end;

  // Line selection
  for I := 0 to High(LineColors) do begin
    if Assigned(Lines.SearchByColor(LineColors[I])) then begin
      Canvas.Brush.Color := LineColors[I];
      Size := 15;
    end else begin
      Canvas.Brush.Color := clSilver;
      Size := 5;
    end;
    Canvas.Pen.Color := clBlack;
    if Assigned(SelectedLine) and (SelectedLine.Color = LineColors[I]) then begin
      Canvas.Pen.Style := psSolid;
    end else begin
      Canvas.Pen.Style := psClear;
    end;

    Canvas.EllipseC(Canvas.Width div 2 - Length(LineColors) div 2 * LineColorsDist + I * LineColorsDist,
      Canvas.Height - LineColorsDist, Size, Size);
  end;

  // Interface
  Text := IntToStr(ServedPassengerCount);
  Canvas.Draw(Canvas.Width - 140, 20, ImagePassenger.Picture.Bitmap);
  Canvas.Brush.Style := bsClear;
  Canvas.TextOut(Canvas.Width - 146 - Canvas.TextWidth(Text), 25, Text);

  Text := IntToStr(Trains.GetUnusedCount);
  Canvas.Draw(Canvas.Width - 240, 20, ImageLocomotive.Picture.Bitmap);
  Canvas.Brush.Style := bsClear;
  Canvas.TextOut(Canvas.Width - 246 - Canvas.TextWidth(Text), 25, Text);

  DrawClock(Canvas);

  // Game over
  if State = gsGameOver then
  begin
    Canvas.Font.Size := 40;
    Canvas.Font.Color := clBlack;
    Canvas.TextOut((Canvas.Width - Canvas.TextWidth(GameOverText)) div 2, 100, GameOverText);
    Canvas.Font.Size := 14;
    Canvas.TextOut((Canvas.Width - Canvas.TextWidth(GameOverReason)) div 2, 160, GameOverReason);
    Text := Format(GameOverStatistic, [ServedPassengerCount, Trunc(Time)]);
    Canvas.TextOut((Canvas.Width - Canvas.TextWidth(Text)) div 2, 180, Text);
  end;
end;

end.

